/********************************************************************************
 * Copyright (c) Des Herriott 1993, 1994
 *               Erik Kunze   1995, 1996
 *               Jens Kilian  1998
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 *         Erik Kunze
 *
 * changed by EKU
 * modified for BeZX JJK
 *******************************************************************************/

static char rcsid[] = "$Id: mem.c,v 3.0 1996/10/03 10:23:45 erik Exp $";

/********************************************************************************
 * mem.c - Where memory is actually held.
 *******************************************************************************/

#include <stdlib.h>
#include <stdio.h>

#include "ZXConfig.h"
#include "ZXMime.h"
#include "z80.h"
#include "io.h"
#include "util.h"
#include "mem.h"
#include "loadsave.h"

/********************************************************************************
 * Defines
 *******************************************************************************/

/********************************************************************************
 * Forwards
 *******************************************************************************/

static int loadRomImage(char *, int);

/********************************************************************************
 * Globals
 *******************************************************************************/

/* Memory is implemented as an array of 4 16K pages. Virtual page numbers range
   from 0..3, real page numbers range from 0..11. <real_pageno> is used to decide
   if the page is read-only memory. */

int RealPageNumbers[4];	/* index to real memory page */
uns8 *TheMemory[4];		/* 4 pointers to real memory */
uns8 *RealMemory[NPAGES];	/* real memory is kept here */
int RomOverride = 0;	/* overwrite ROM in write_byte and write_word */

/********************************************************************************
 * Locals
 *******************************************************************************/

/********************************************************************************
 * InitMemory() - Allocate NPAGES chunks of memory, each of size XZX_PAGESIZE.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
InitMemory(void)
{
	static int inited = 0;
	int i;

	if (inited)
	{
		XzxMesg(XZX_ERR, "attempt to re-initialise memory (failed)");
		return;
	}

	for (i = 0; i < NPAGES; i++)
	{
		if (!(RealMemory[i] = (uns8 *)malloc(XZX_PAGESIZE)))
		{
			XzxMesg(XZX_FATAL, "can't allocate 16K for page %d", i);
		}
	}
}

/********************************************************************************
 * InitPages() - Set up the initial memory pages; on all machines, we page in
 *               ROM0 at 0x0000, RAM5 at 0x4000, RAM2 at 0x8000 and RAM0 at
 *               0xc000.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
InitPages(void)
{
	PageIn(0, ROM0);
	PageIn(1, RAM5);
	PageIn(2, RAM2);
	PageIn(3, RAM0);
}

/********************************************************************************
 * PageIn() -  Page in real memory page at <frame>, where <frame> is in range
 *             0..3.
 *
 * Input:  frame - # of frame (0-3)
 *         rpage - # of page (ROM0-NPAGES)
 *
 * Return: -
 *******************************************************************************/

void
PageIn(int frame, int rpage)
{
	/* Only the +3 allows RAM at all 4 frames */
	if (GETCFG(machine) != 3 && !frame && rpage >= RAM0)
	{
		XzxMesg(XZX_WARN, "attempt to page RAM in at 0x0000 (ignored)");
		return;
	}

	TheMemory[frame] = RealMemory[rpage];
	RealPageNumbers[frame] = rpage;

	/* If paging in ROM on the +3, also page in RAM 5, 2, Last0x7FFD */
	if (GETCFG(machine) == 3 && !frame && rpage < RAM0)
	{
		PageIn(1, RAM5);
		PageIn(2, RAM2);
		PageIn(3, RAM0 + (Last0x7FFD & BIT_SELRAM));
	}
}

/********************************************************************************
 * loadRomImage() - Read the ROM image <image> into real memory page <page>.
 *
 * Input:  image - name of image file
 *         page  - page to load in the image
 *
 * Return: 0 - o.k.
 *         1 - error
 *******************************************************************************/

static int
loadRomImage(char *image, int page)
{
	FILE *ifp;
	size_t nread;

	if (!(ifp = OpenXzxFile(image, "r", ZXMimeTypes::kROMFile)))
	{
		XzxMesg(XZX_PERR, "can't open ROM image <%s>", image);
		return 1;
	}

	nread = fread(RealMemory[page], 1, XZX_PAGESIZE, ifp);

	/* Give a warning if image is not 16K or 8K */
	if (nread != XZX_PAGESIZE && nread != XZX_PAGESIZE / 2)
	{
		XzxMesg(XZX_WARN, "ROM image <%s> is only %d bytes", image, nread);
	}

	fclose(ifp);

	return 0;
}

/********************************************************************************
 * ReadROMs() - Load rom images from files into memory.
 *
 * Input:  -
 *
 * Return:  0 - o.k.
 *         -1 - error
 *******************************************************************************/

int
ReadROMs(void)
{
	int ret = 0;
	
	switch(GETCFG(machine))
	{
		case 48:
			ret |= loadRomImage(GETCFG(rom48), ROM0);
			break;
		case 128:
			ret |= loadRomImage(GETCFG(rom128_0), ROM0);
			ret |= loadRomImage(GETCFG(rom128_1), ROM1);
			break;
		case 3:
			ret |= loadRomImage(GETCFG(rompl3_0), ROM0);
			ret |= loadRomImage(GETCFG(rompl3_1), ROM1);
			ret |= loadRomImage(GETCFG(rompl3_2), ROM2);
			ret |= loadRomImage(GETCFG(rompl3_3), ROM3);
			break;
	}

	if (ret)
	{
		/* Error in loading ROM images */
		return -1;
	}

	SetLoadSaveTraps();

	/* No support for Interface 1 on the +3 */
	if (GETCFG(if1_active) && GETCFG(machine) != 3)
	{
		if (loadRomImage(GETCFG(if1_rom), ROM2) == -1)
		{
			XzxMesg(XZX_WARN, "IF1 emulation disabled");
			SETCFG(if1_active, false);
		}
	}

	/* No support for Multiface 128 on the +3 */
	if (GETCFG(mf128_active) && GETCFG(machine) != 3)
	{
		if (loadRomImage(GETCFG(mf128_rom), ROM3) == -1)
		{
			XzxMesg(XZX_WARN, "MF128 emulation disabled");
			SETCFG(mf128_active, false);
		}
	}

	return 0;
}

/********************************************************************************
 * SelectModel() - Select a Spectrum model to emulate.
 *
 * Input:  model - 47/128/3
 *         reset - reset the CPU
 *
 * Return:  0 - o.k.
 *         -1 - error
 *******************************************************************************/

int
SelectModel(int model, int reset)
{
	if (model == 48 || model == 128 || model == 3)
	{
		if (GETCFG(machine) != model)
		{
			SETCFG(machine, model);
			BasicRom = GET_BASIC_ROM(model);

			if (reset)
			{
				/* This does a InitPages() implicitly */
				Z80_Reset();
			}
			else
			{
				InitPages();
			}

			if (ReadROMs() != -1)
			{
				return 0;
			}
		}
	}
	else
	{
		XzxMesg(XZX_ERR, "no such memory model %d", model);
	}

	return -1;
}

/* EOF *************************************************************************/
