/********************************************************************************
 * Copyright (c) Des Herriott 1993, 1994
 *               Erik Kunze   1995, 1996
 *               Jens Kilian  1998
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 *         Erik Kunze
 *
 * changed by EKU
 * modified for BeZX JJK
 *******************************************************************************/

static char rcsid[] = "$Id: screen.c,v 3.1 1996/10/03 19:08:32 erik Exp $";

/********************************************************************************
 * screen.c - Display file handling.
 *******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Alert.h>
#include <Bitmap.h>
#include <OS.h>
#include <Rect.h>
#include <Screen.h>
#include <View.h>

#include "ZXConfig.h"
#include "ZXFileDialog.h"
#include "ZXView.h"
#include "z80.h"
#include "debug.h"
#include "mem.h"
#include "util.h"
#include "audio.h"
#include "fdc.h"
#include "if1.h"
#include "main.h"
#include "snapshot.h"
#include "tables.h"
#include "version.h"
#include "screen.h"

/********************************************************************************
 * Defines
 *******************************************************************************/

#define bcopy(s,d,l)	memcpy(d,s,l)
#define bzero(s,l)		memset(s,l,0)

#define shade(x,inkpap,paper) \
						(((x) & (inkpap)) ^ (paper))

/* Map an (x,y) pixel coordinate to a offset into the attribute area */
#define XYToAttr(x,y)	(((x) >> 3) + ((y & 0xf8) << 2))

/* Extend the bounding box if necessary */
#define	EXPAND_BOX	{ \
						if (x < x_min) x_min = x;      \
						x += (GETCFG(scale) << 3) - 1; \
						if (x > x_max) x_max = x;      \
						if (y < y_min) y_min = y;      \
						y += GETCFG(scale) - 1;        \
						if (y > y_max) y_max = y;      \
						screenDirty++;                 \
					}

/* Extend the bounding box to maximum size */
#define MAXIMIZE_BOX	{ \
						x_min = y_min = 0; \
						x_max = hsize - 1; \
						y_max = vsize - 1; \
						screenDirty++;     \
					}

/* Reset the bounding box */
#define MINIMIZE_BOX	{ \
						x_max = y_max = 0;					   \
						x_min = hsize - 1;					   \
						y_min = vsize - 1;					   \
						screenDirty = 0;					   \
						bzero(scrChanged, sizeof(scrChanged)); \
					}

/********************************************************************************
 * Forwards
 *******************************************************************************/

static void screenWrite(uns16, uns8);
static void quitScreen(void);
static void getColors(uns8 *);

/********************************************************************************
 * Globals
 *******************************************************************************/

/********************************************************************************
 * Locals
 *******************************************************************************/

static ZXView *pView;

/* Common variables */
static uns8 white, black;
static uns8 *pDisplayBits;
static int bytesPerLine;
static int hsize, vsize;
static void **bmAddrTable;
static char scrChanged[PIXEL_LENGTH];

/* RGB triples describing the Spectrum's 16 colors.  Note that black and bright
   black are the same. */
static unsigned short rgbVals[NCOLORS][3] = {
	{ 0x0000, 0x0000, 0x0000 },			/* black		  */
	{ 0x0000, 0x0000, 0xbfff },			/* blue			  */
	{ 0xbfff, 0x0000, 0x0000 },			/* red			  */
	{ 0xbfff, 0x0000, 0xbfff },			/* magenta		  */
	{ 0x0000, 0xbfff, 0x0000 },			/* green		  */
	{ 0x0000, 0xbfff, 0xbfff },			/* cyan			  */
	{ 0xbfff, 0xbfff, 0x0000 },			/* yellow		  */
	{ 0xbfff, 0xbfff, 0xbfff },			/* white		  */
	{ 0x0000, 0x0000, 0x0000 },			/* black		  */
	{ 0x0000, 0x0000, 0xffff },			/* bright blue	  */
	{ 0xffff, 0x0000, 0x0000 },			/* bright red	  */
	{ 0xffff, 0x0000, 0xffff },			/* bright magenta */
	{ 0x0000, 0xffff, 0x0000 },			/* bright green	  */
	{ 0x0000, 0xffff, 0xffff },			/* bright cyan	  */
	{ 0xffff, 0xffff, 0x0000 },			/* bright yellow  */
	{ 0xffff, 0xffff, 0xffff },			/* bright white	  */
};

static uns8 colors[NCOLORS];	/* 16 Spectrum colors */
static uns8 ink[FLASH];
static uns8 paper[FLASH];

static int x_min, x_max, y_min, y_max;	/* bounding box of last area changed */
static int screenDirty;					/* does the screen need refreshing? */
static int flashPhase;					/* gets toggled between 1 and 0 twice
										   per second */

/********************************************************************************
 * screenWrite() - Write to the screen.
 *
 * Input:  addr - memory address
 *         val  - byte
 *
 * Return: -
 *******************************************************************************/

static void
screenWrite(uns16 addr, uns8 val)
{
	int x, y;
	uns8 *bmptr;
	uns8 ipix, ppix;
	uns8 attr;
	int j, k;
	uns8 pmask = 0, ipmask = 0;

	addr &= 0x3fff;
	x = DtblX[addr];
	y = DtblY[addr];
	attr = RealMemory[ScreenSelect][ATTR_OFFSET + XYToAttr(x,y)];
	
	/* Toggle the colors if we're in the inverse part of the FLASH cycle */
	if (flashPhase && attr & FLASH) {
		ipix = paper[attr & ~FLASH];
		ppix = ink[attr & ~FLASH];
	} else {
		ipix = ink[attr & ~FLASH];
		ppix = paper[attr & ~FLASH];
	}

	x *= GETCFG(scale);
	y *= GETCFG(scale);

	bmptr = (uns8 *)bmAddrTable[addr];
	for (j = 0; j < 8; j++) {
		for (k = 0; k < GETCFG(scale); k++) {
			*bmptr++ = val & 0x80 ? ipix : ppix;
		}
		val <<= 1;
	}
	for (j = 1; j < GETCFG(scale); j++) {
		bcopy((uns8 *)bmAddrTable[addr],
			  (uns8 *)bmAddrTable[addr] + j * bytesPerLine,
			  (sizeof(uns8) << 3) * (size_t)GETCFG(scale));
	}

	EXPAND_BOX;
}

/********************************************************************************
 * ScreenWrite() - This one is similar to a wrapper. Don't write actually to
 *                 the screen, but remember the change. This is for those
 *                 'rainbowing' programs, that will eat all the CPU time without
 *                 it.
 *
 * Input:  addr - memory address
 *         val  - byte
 *
 * Return: -
 *******************************************************************************/

void
ScreenWrite(uns16 addr, uns8 /*val*/)
{
	scrChanged[addr & 0x3fff] = '\001';
	screenDirty++;
}

/********************************************************************************
 * ScreenRefresh() - Refresh screen.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void 
ScreenRefresh(void)
{
	if (screenDirty) {
		int i;

		for (i = 0; i < PIXEL_LENGTH; i++) {
			if (scrChanged[i]) {
				screenWrite(i, RealMemory[ScreenSelect][i]);
			}
		}

		BRect rect(x_min, y_min, x_max, y_max);
		pView->Refresh(rect);
		MINIMIZE_BOX;
	}
}

/********************************************************************************
 * ForceScreenRefresh() -  For 128K and +3 machines, if the screen select bit
 *                         is toggled, force an update of the screen to
 *                         reflect the new display file.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
ForceScreenRefresh(void)
{
	int i;
	uns8 *mem = &RealMemory[ScreenSelect][0];

	for (i = DFILE; i < DFILE + PIXEL_LENGTH; i++, mem++) {
		ScreenWrite(i, *mem);
	}
}

/********************************************************************************
 * AttrWrite() - <val> has just been written to <addr> in the attribute area,
 *               update the screen with ScreenWrite().
 *
 * Input:  addr - memory address
 *         val  - byte
 *
 * Return: -
 *******************************************************************************/

void
AttrWrite(uns16 addr, uns8 /*val*/)
{
	int i;
	uns8 *mem;

	addr = ((addr & 0x300) << 3) + (addr & 0xff);
	mem = &RealMemory[ScreenSelect][addr];

	for (i = 0; i < 8; i++, addr += X_PIXELS, mem += X_PIXELS) {
		ScreenWrite(addr, *mem);
	}
}

/********************************************************************************
 * DoFlashing() - Any byte in the attribute area with the FLASH bit set has its
 *                ink and paper swapped twice/sec.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
DoFlashing(void)
{
	int i;
	uns8 *mem;
	uns16 addr;

	if (GETCFG(flashing)) {
		flashPhase = !flashPhase;

		addr = DFILE + ATTR_OFFSET;
		mem = &RealMemory[ScreenSelect][ATTR_OFFSET];

		for (i = 0; i < ATTR_LENGTH; i++, mem++, addr++) {
			if (*mem & FLASH) {
				AttrWrite(addr, *mem);
			}
		}
	}
}

/********************************************************************************
 * SetBorderColor() - Set the Spectrum's border color.
 *
 * Input:  color - border color
 *
 * Return: -
 *******************************************************************************/

void
SetBorderColor(int color)
{
	uint8 red, green, blue;

	if (GETCFG(mono)) {
		red = green = blue = color < 4 ? 0 : 255;
	} else {
		red   = rgbVals[color][0] >> 8;
		green = rgbVals[color][1] >> 8;
		blue  = rgbVals[color][2] >> 8;
	}
	pView->SetBorderColor(red, green, blue);
}

/********************************************************************************
 * quitScreen() - Perform a clean shutdown.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

static void
quitScreen(void)
{
	if (bmAddrTable) {
		free(bmAddrTable);
	}
	
	// Notify UI that emulator is quitting.
	pView->Window()->PostMessage(B_QUIT_REQUESTED);
}

/********************************************************************************
 * initImage() - Image initialisation routine.
 *
 * Input:  -
 *
 * Return:  0 - o.k.
 *         -1 - error
 *******************************************************************************/

static int
initImage(void)
{
	BBitmap *pBuffer;
	void **ptr;
	int x, y, z;
	
	pBuffer = pView->Buffer();
	pDisplayBits = (uns8 *)(pBuffer->Bits());

	bytesPerLine = pBuffer->BytesPerRow();
	XDEB(fprintf(stderr, "bytesPerLine : %d\n", bytesPerLine);)
	
	bmAddrTable = 0;
	if (bytesPerLine == 256 * GETCFG(scale)) {
		/* Create a table to convert from Spectrums screen addresses to BBitmap */
		if (!(bmAddrTable = (void **)malloc(sizeof(void *) * PIXEL_LENGTH))) {
			XzxMesg(XZX_FATAL, "malloc failed in initImage()");
		}

		ptr = bmAddrTable;
		for (z = 0; z < PIXEL_LENGTH; z++) {
			x = DtblX[z];
			y = DtblY[z];
			*ptr++ = pDisplayBits + GETCFG(scale) * (x + y * hsize);
		}

	} else {
		XzxMesg(XZX_FATAL, "%d bytes per line not supported", bytesPerLine);
	}

	return 0;
}

/********************************************************************************
 * getColors() - Initialise the 16 colors required by the Speccy, or just black
 *               and white if monochrome is set.
 *
 * Input:  col - pointer to pixel color return value
 *
 * Return: -
 *******************************************************************************/

static void
getColors(uns8 *col)
{
	BScreen screen;
	int i, j, k;

	white = screen.IndexForColor(255, 255, 255);
	black = screen.IndexForColor(0, 0, 0);

	if (GETCFG(mono)) {
		for (i = 0; i < 8; i++) {
			col[i] = col[i + 8] = i < 4 ? black : white;
		}

	} else {
		for (i = 0; i < NCOLORS; i++) {
			uint8 red   = rgbVals[i][0] >> 8;
			uint8 green = rgbVals[i][1] >> 8;
			uint8 blue  = rgbVals[i][2] >> 8;
			col[i] = screen.IndexForColor(red, green, blue);
			
			// Ensure consistency with border color.
			rgb_color color = screen.ColorForIndex(col[i]);
			rgbVals[i][0] = color.red   << 8;
			rgbVals[i][1] = color.green << 8;
			rgbVals[i][2] = color.blue  << 8;
		}
	}

	/* Build hash tables to speed up scrWriteXXX() */
	if (GETCFG(mono)) {
		for (i = 0; i < 2; i ++) {
			for (j = 0; j < 8; j++) {
				for (k = 0; k < 8; k++) {
					if (j > k) {
						ink  [(i << 6) + (j << 3) + k] = black;
						paper[(i << 6) + (j << 3) + k] = white;

					} else if (j < k) {
						ink  [(i << 6) + (j << 3) + k] = white;
						paper[(i << 6) + (j << 3) + k] = black;
						
					} else {
						if (j < GREEN || (j > WHITE && j < BRIGHT_GREEN)) {
							ink  [(i << 6) + (j << 3) + k] = black;
							paper[(i << 6) + (j << 3) + k] = black;

						} else {
							ink  [(i << 6) + (j << 3) + k] = white;
							paper[(i << 6) + (j << 3) + k] = white;
						}
					}
				}
			}
		}

	} else /* !GETCFG(mono) */ {
		for (i = 0; i < 2; i ++) {
			for (j = 0; j < 8; j++) {
				for (k = 0; k < 8; k++) {
					ink  [(i << 6) + (j << 3) + k] = col[k + (i << 3)];
					paper[(i << 6) + (j << 3) + k] = col[j + (i << 3)];
				}
			}
		}
	}
}

/********************************************************************************
 * InitScreen() - Initialize display handling.
 *
 * Input:  view - reference to the ZXView object
 *
 * Return: -
 *******************************************************************************/

void
InitScreen(ZXView &view)
{
	/* Initialize global variables. */
	pView = &view;
	hsize = HSize;
	vsize = VSize;

	getColors(colors);
	SetBorderColor(WHITE);
	flashPhase = !GETCFG(flashing);

	/* Image handling initialisation */
	initImage();
	MINIMIZE_BOX;

	OnQuit(quitScreen);
}

/********************************************************************************
 * CheckEvents() - Check and process any incoming asynchronous events.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
CheckEvents(void)
{
	int32 msgCode;
	while (read_port_etc(pView->EventPort(),&msgCode,0,0,B_TIMEOUT,0) >= B_OK) {
		// Process the event.
		switch (msgCode) {

		case ZXView::kEmulatorReset:
			Z80_Reset();
			break;

		case ZXView::kEmulatorNMI:
			Z80_NMI();
			break;

		case ZXView::kSaveSnapshot:
			{	char *pStr = Dialog("Name of snapshot to save", true);
				if (pStr) {
					WriteSnapshot(pStr);
				}
			}
			break;

		case ZXView::kLoadSnapshot:
			{	char *pStr = Dialog("Name of snapshot to load", false);
				if (pStr) {
					ReadSnapshot(pStr);
				}
			}
			break;

		case ZXView::kAudioToggle:
//			SpeakerOnOff();
//			AyOnOff();
			break;

		case ZXView::kDebugToggle:
#ifdef DEBUG
			Debug = !Debug;
#endif
			break;

		case ZXView::kInsertDiskA:
			if (GETCFG(machine) == 3) {
				char *pStr = Dialog("Disk image for drive A", false);
				if (pStr) {
					InsertDisk(pStr, 'A');
				}
			}
			break;

		case ZXView::kInsertDiskB:
			if (GETCFG(machine) == 3) {
				char *pStr = Dialog("Disk image for drive B", false);
				if (pStr) {
					InsertDisk(pStr, 'B');
				}
			}
			break;

		case ZXView::kSelectModel:
			switch ((new BAlert("", "Select memory model.", "48", "128", "+3",
									  B_WIDTH_AS_USUAL, B_EMPTY_ALERT))->Go()) {
			case 0:
				SelectModel(48, true);
				break;

			case 1:
				SelectModel(128, true);
				break;

			case 2:
				SelectModel(3, true);
				break;
			}
			break;

		case ZXView::kScreenDump:
			{	char *pStr = Dialog("Name of screen dump to save", true);
				if (pStr) {
					WriteScreenDump(pStr);
				}
			}
			break;

		case ZXView::kEmulatorQuit:
			Quit(0);
			break;
		}
	}
}

/********************************************************************************
 * Dialog() - Pop up a file selection dialog and wait until a file has been
 *            selected.  Return the selected file name.
 *
 * Input:  pLabel - window label
 *         forSaving - true iff file is for output
 *
 * Return: str - input string
 *         0   - no valid input
 *******************************************************************************/

char *
Dialog(char *pLabel, bool forSaving)
{
	static char *pResult = 0;
	if (pResult) {
		free(pResult);
	}
	pResult =
		(new ZXFileDialog(forSaving ? B_SAVE_PANEL : B_OPEN_PANEL, pLabel))->Go();
	
	return (pResult && *pResult) ? pResult : 0;
}

/* EOF *************************************************************************/
